#include "bpop_main.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

int playSounds = 1;
GameScene currentScene = GAME_SCENE_MENU;
GameScene nextScene = GAME_SCENE_NONE;
PlayfieldScores finalScores;
Sound sfxPop;
GameDifficulty gameDifficulty = GAME_DIFFICULTY_NORMAL;

Storage storage = {
    .version = 1,
};

Texture2D texShrimpy;
Texture2D texLeftCorner;
Texture2D texUiAtlas;
Texture2D texGround;
Font fontMainMedium;
Font fontMainLarge;

NPatchInfo nPatchButton = {
    .source = {0,0,240,145},
    .left = 10,
    .top = 10,
    .right = 10,
    .bottom = 10,
    .layout = NPATCH_NINE_PATCH
};
Rectangle rectUiAtlasCircle = {
    .x = 250,
    .y = 0,
    .width = 148,
    .height = 148,
};
Rectangle rectUiAtlasShrimpy = {
    .x = 420,
    .y = 390,
    .width = 512-420,
    .height = 512-390,
};

int outlineSizeLoc = 0;
int outlineColorLoc = 0;
int textureSizeLoc = 0;

Camera3D camera = { 
    .position = { 0.0f, 0.0f, -10.0f },
    .target = { 0.0f, 0.0f, 0.0f },
    .up = { 0.0f, 1.0f, 0.0f },
    .projection = CAMERA_ORTHOGRAPHIC,
    .fovy = 320.0f,
};


Playfield playfield = {
    .fieldSize = {250, 300},
    .waterLevel = 200.0f,
    .minHeight = 90.0f,
    .maxHeight = 280.0f,
};

GameTime gameTime = {
    .fixedDeltaTime = 1.0f / 60.0f,
};

Color bubbleTypeColors[] = {
    COLOR_FROM_HEX(0xaaccff),
    COLOR_FROM_HEX(0x2277ff),
    COLOR_FROM_HEX(0x33ffaa),
    COLOR_FROM_HEX(0x9933aa),
    COLOR_FROM_HEX(0xaa33ff),
};

Model bubbleModel;
Shader bubbleOutlineShader;
RenderTexture2D bubbleFieldTexture;

int isClickActionBlocked = 0;

PlayfieldScores CalculatePlayfieldScores(Playfield *playfield);
void SaveStorage();

void DrawText2(const char *text, int x, int y, Color color)
{
    DrawTextEx(fontMainMedium, text, (Vector2){x, y}, fontMainMedium.baseSize, 1, color);
}

void DrawText3(const char *text, int x, int y, Color color)
{
    DrawTextEx(fontMainLarge, text, (Vector2){x, y}, fontMainMedium.baseSize, 1, color);
}

float MeasureText2(const char *text)
{
    return MeasureTextEx(fontMainMedium, text, fontMainMedium.baseSize, 1).x;
}

float BubbleLevelRadius(int level)
{
    return powf((level + 1) * 120, .5f);
}

int IsClickActioned()
{
    int result = 0;
    if (IsMouseButtonReleased(MOUSE_LEFT_BUTTON) && !isClickActionBlocked)
    {
        result = 1;
    }
    return result;
}

int Button(const char *text, Vector2 position, Vector2 size)
{
    Rectangle rect = {position.x, position.y, size.x, size.y};
    int isHovered = !isClickActionBlocked && CheckCollisionPointRec(GetMousePosition(), rect);
    int result = isHovered && IsClickActioned();

    Color color = isHovered ? COLOR_FROM_HEX(0xcceeff) : COLOR_FROM_HEX(0xaaccff);
    if (!isClickActionBlocked && isHovered && IsMouseButtonDown(MOUSE_LEFT_BUTTON))
    {
        color = COLOR_FROM_HEX(0xddf8ff);
    }

    DrawTextureNPatch(texUiAtlas, nPatchButton, rect, Vector2Zero(), 0.0f, color);

    int fontSize = fontMainMedium.baseSize;
    Vector2 tsize = MeasureTextEx(fontMainMedium, text, fontSize, 1);
    DrawTextEx(fontMainMedium, text, (Vector2){position.x + size.x * 0.5f - tsize.x * 0.5f, position.y + size.y * 0.5f - tsize.y * 0.5f}, fontSize, 1, BLACK);
    return result;
}

void PlayfieldFixedUpdate(Playfield *playfield, int popBubbles)
{
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        bubble->sameLevelContact = 0;
    }

    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (!bubble->flagIsActive) continue;
        float r = bubble->radius;

        for (int j = i + 1; j < MAX_BUBBLES; j++)
        {
            Bubble *other = &playfield->bubbles[j];
            if (!other->flagIsActive) continue;
            float otherR = other->radius;
            float sumR2 = (r + otherR) * (r + otherR);
            float d2 = Vector2DistanceSqr(bubble->position, other->position);
            int canMerge = bubble->bubbleLevel == other->bubbleLevel && bubble->bubbleType == other->bubbleType;
            if (d2 < sumR2 * 1.05f)
            {
                if (canMerge)
                {
                    bubble->sameLevelContact = 1;
                    other->sameLevelContact = 1;
                }
                if (canMerge && bubble->bubbleMergeCooldown <= 0.0f 
                    && other->bubbleMergeCooldown <= 0.0f)
                {
                    // merge bubbles
                    bubble->bubbleLevel++;
                    bubble->position = Vector2Lerp(bubble->position, other->position, 0.5f);
                    bubble->radius = BubbleLevelRadius(bubble->bubbleLevel);
                    bubble->bubbleMergeCooldown = 1.0f;

                    playfield->poppingBubbles[i] = *bubble;
                    playfield->poppingBubbles[i].lifeTime = 0.0f;
                    playfield->comboPopPoints += playfield->comboPop;
                    playfield->comboPop++;

                    if (playSounds)
                        PlaySound(sfxPop);

                    other->flagIsActive = 0;
                }
            }

            if (d2 < sumR2)
            {
                float overlap = r + otherR - sqrtf(d2);
                Vector2 normal = Vector2Normalize(Vector2Subtract(other->position, bubble->position));
                // resolve overlap by moving the bubbles apart
                const float errorCorrection = 0.25f;
                bubble->position = Vector2Subtract(bubble->position, Vector2Scale(normal, overlap * errorCorrection));
                other->position = Vector2Add(other->position, Vector2Scale(normal, overlap * errorCorrection));

                // bounce off each other
                Vector2 relativeVelocity = Vector2Subtract(bubble->velocity, other->velocity);
                float dot = Vector2DotProduct(relativeVelocity, normal);
                if (dot > 0.0f)
                {
                    // DrawLineV(bubble->position, other->position, COLOR_FROM_HEX(0xff0000));
                    float impulse = -dot * 0.85f;
                    bubble->velocity = Vector2Add(bubble->velocity, Vector2Scale(normal, impulse));
                    other->velocity = Vector2Subtract(other->velocity, Vector2Scale(normal, impulse));
                }
            }
        }

        if (!bubble->sameLevelContact)
        {
            bubble->bubbleMergeCooldown = 1.0f;
        }
        else
        {
            bubble->bubbleMergeCooldown -= gameTime.fixedDeltaTime;
        }

        float buoyancy = -20.0f;
        if (bubble->position.y < playfield->waterLevel)
        {
            buoyancy = (playfield->waterLevel - bubble->position.y) * 0.5f;
        }

        int isOutsideZone = bubble->position.y < playfield->minHeight + bubble->radius || 
            bubble->position.y > playfield->maxHeight - bubble->radius;
        bubble->lifeTime += gameTime.fixedDeltaTime;
        bubble->isOutsideZone = isOutsideZone && bubble->lifeTime > 1.0f;

        float horizontalV = 0.0f;
        if (playfield->enableTutorialMerge) {
            horizontalV = fminf(fmaxf(-10.0f, (playfield->fieldSize.x / 2 - bubble->position.x) * 0.1f), 10.0f);
        }
        bubble->velocity = Vector2Add(bubble->velocity, (Vector2){horizontalV, buoyancy});
        bubble->velocity = Vector2Scale(bubble->velocity, 0.92f);
        bubble->position.x += bubble->velocity.x * gameTime.fixedDeltaTime;
        bubble->position.y += bubble->velocity.y * gameTime.fixedDeltaTime;
        if ((bubble->position.x < r && bubble->velocity.x < 0.0f) || 
            (bubble->position.x > playfield->fieldSize.x - r && bubble->velocity.x > 0.0f))
        {
            bubble->velocity.x *= -0.9f;
        }

        bubble->position.x = (bubble->position.x < r) ? r : (bubble->position.x > playfield->fieldSize.x - r) ? playfield->fieldSize.x - r : bubble->position.x;

        if (popBubbles && GetRandomValue(0, 1000) < 50)
        {
            playfield->poppingBubbles[i] = *bubble;
            playfield->poppingBubbles[i].lifeTime = 0.0f;
            bubble->flagIsActive = 0;
        }
        // bubble->position.y = (bubble->position.y < r) ? r : (bubble->position.y > playfield->fieldSize.y - r) ? playfield->fieldSize.y - r : bubble->position.y;

        // debug velocity
        // DrawLineV(bubble->position, Vector2Add(bubble->position, Vector2Scale(bubble->velocity, 1.0f)), COLOR_FROM_HEX(0xff0000));
    }

    int outsideCount = 0;
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (bubble->isOutsideZone)
        {
            outsideCount++;
        }
    }

    if (outsideCount == 0)
    {
        playfield->strikes = 0;
        playfield->gameoverCooldown = 0.0f;
    }
    if (playfield->strikes >= MAX_STRIKES && !popBubbles)
    {
        playfield->gameoverCooldown += gameTime.fixedDeltaTime;
        if (playfield->gameoverCooldown > GAMEOVER_COOLDOWN)
        {
            finalScores = CalculatePlayfieldScores(playfield);
            nextScene = GAME_SCENE_GAMEOVER;
        }
    }
}

void PlayfieldTryAddBubble(Playfield *playfield, Vector2 position)
{
    if (playfield->strikes >= MAX_STRIKES) return;

    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (!bubble->flagIsActive)
        {
            bubble->flagIsActive = 1;
            bubble->position = position;
            bubble->velocity = (Vector2){GetRandomValue(-100, 100), GetRandomValue(-100, 100)};
            bubble->bubbleType = playfield->nextBubbleType;
            bubble->bubbleLevel = playfield->nextBubbleLevel;
            bubble->radius = BubbleLevelRadius(bubble->bubbleLevel);
            bubble->lifeTime = 0.0f;
            playfield->spawnedBubbleCount++;

            for (int j = 0; j < MAX_BUBBLES; j += 1)
            {
                Bubble *other = &playfield->bubbles[j];
                if (!other->flagIsActive) continue;
                if (other->isOutsideZone)
                {
                    playfield->strikes++;
                    break;
                }
            }
            playfield->comboPop = 0;
            playfield->nextBubbleType = GetRandomValue(0, gameDifficulty > 0 ? gameDifficulty : 1);
            playfield->nextBubbleLevel = GetRandomValue(0, 3);
            break;
        }
    }

}

PlayfieldScores CalculatePlayfieldScores(Playfield *playfield)
{
    PlayfieldScores scores = {0};
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (bubble->flagIsActive)
        {
            scores.bubbleCount++;
            uint32_t bubbleScore = 1 << bubble->bubbleLevel;
            scores.score += bubbleScore;

            if (bubble->isOutsideZone)
            {
                scores.outsideBubbleCount++;
            }
        }
    }
    scores.score += playfield->spawnedBubbleCount;
    scores.score += playfield->comboPopPoints;
    return scores;
}

Vector2 PlayfieldPositionToSpawnPosition(Playfield *playfield, Vector2 position)
{
    Vector2 spawnPosition = position;
    spawnPosition.y = BubbleLevelRadius(5);
    return spawnPosition;
}

Vector2 PlayfieldScreenToSpawnPosition(Playfield *playfield, Camera3D camera, Vector2 screenPosition)
{
    Vector3 cursorPosition = GetScreenToWorldRay(screenPosition, camera).position;
    cursorPosition.x += playfield->fieldSize.x / 2;
    cursorPosition.y += playfield->fieldSize.y / 2;

    Vector2 pos = {cursorPosition.x, cursorPosition.y};
    return PlayfieldPositionToSpawnPosition(playfield, pos);
}

void DrawBubblePop(Vector3 position, float t, int level, Color color)
{
    // int oldSeed = GetRandomValue(0, 0xfffffff);
    // SetRandomSeed((int)(position.x * 1000) ^ (int)(position.y * 1000) + level * 1000);
    color.a = 255 * (1.0f - t);
    DrawModel(bubbleModel, position, BubbleLevelRadius(level) * (1 + t), color);
}

void DrawBubble(Vector3 position, int level, Color color)
{
    float bubbleExtraRadius = 5.0f;
    float r = BubbleLevelRadius(level) + bubbleExtraRadius;
    DrawModel(bubbleModel, position, r, color);
    if (level < 1) return;
    position.z -= r;
    float tinyR = level < 6 ? 2 : 4;
    int count = level < 6 ? level : level - 5;
    for (int i = 0; i < count; i++)
    {
        float ang = (i * 25.0f + 30.0f) * DEG2RAD;
        float offsetR = i % 2 == 0 ? 0.4f : 0.7f;
        Vector3 offset = {cosf(ang) * offsetR * r, sinf(ang) * offsetR * r, 0};
        DrawModel(bubbleModel, Vector3Add(position, offset), tinyR, WHITE);
    }
}

void PlayfieldDrawBubbles(Playfield *playfield, Camera3D camera)
{
    DrawCube((Vector3){0, 0, 0}, playfield->fieldSize.x, playfield->fieldSize.y, 0, COLOR_FROM_HEX(0xbbddff));
    DrawCube((Vector3){0, (playfield->waterLevel - playfield->fieldSize.y) * 0.5f, 0}, playfield->fieldSize.x, 
        playfield->waterLevel, 0, COLOR_FROM_HEX(0x225588));
    
    // cursor bubble
    if (currentScene == GAME_SCENE_PLAY)
    {
        Vector2 mousePos = GetMousePosition();
        Vector2 spawnPosition = PlayfieldScreenToSpawnPosition(playfield, camera, mousePos);
        Vector3 drawPos = (Vector3){spawnPosition.x - playfield->fieldSize.x * 0.5f, spawnPosition.y - playfield->fieldSize.y * 0.5f, 0};
        if (playfield->strikes < MAX_STRIKES && drawPos.x >= -playfield->fieldSize.x * 0.5f && drawPos.x <= playfield->fieldSize.x * 0.5f)
        {
            DrawBubble(drawPos, playfield->nextBubbleLevel, bubbleTypeColors[playfield->nextBubbleType]);
        }
    }

    // DrawRectangle(0, 0, playfield->fieldSize.x, playfield->fieldSize.y, COLOR_FROM_HEX(0x225588));
    rlPushMatrix();
    rlTranslatef(-playfield->fieldSize.x / 2, -playfield->fieldSize.y / 2, 0);
    // draw bubbles into playfield space
    float blink = sinf(gameTime.time * 10.0f) * 0.2f + 0.5f;
    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->bubbles[i];
        if (!bubble->flagIsActive) continue;
        Vector3 position = {bubble->position.x, bubble->position.y, 0};
        Color bubbleColor = bubbleTypeColors[bubble->bubbleType];
        int isOutsideZone = bubble->isOutsideZone;
        
        if (isOutsideZone)
        {
            bubbleColor = ColorLerp(bubbleColor, COLOR_FROM_HEX(0xff4433), blink);
        }
        // lazy: encode id into rgb values
        bubbleColor.r = bubbleColor.r - i * 2 % 4;
        bubbleColor.g = bubbleColor.g - (i * 2 / 4) % 4;
        bubbleColor.b = bubbleColor.b - (i * 2 / 16);
        bubbleColor.a = 255;
        DrawBubble(position, bubble->bubbleLevel, bubbleColor);
    }

    for (int i = 0; i < MAX_BUBBLES; i++)
    {
        Bubble *bubble = &playfield->poppingBubbles[i];
        if (!bubble->flagIsActive) continue;
        bubble->lifeTime += gameTime.deltaTime;
        float t = bubble->lifeTime / 0.125f;
        if (t >= 1.0f)
        {
            bubble->flagIsActive = 0;
            continue;
        }
        Vector3 position = {bubble->position.x, bubble->position.y, 0};
        DrawBubblePop(position, t, bubble->bubbleLevel, bubbleTypeColors[bubble->bubbleType]);
    }
    rlPopMatrix();
}

void PlayfieldDrawRange(Playfield *playfield, Camera3D camera)
{
    Color rangeLimitColor = COLOR_FROM_HEX(0xff4400);
    int divides = 10;
    float divWidth = playfield->fieldSize.x / divides;
    for (int i = 0; i < divides; i+=2)
    {
        float x = i * divWidth - playfield->fieldSize.x * 0.5f + divWidth * 1.0f;
        DrawCube((Vector3){x, playfield->minHeight - playfield->fieldSize.y * 0.5f, 0}, divWidth, 5, 5, rangeLimitColor);
        DrawCube((Vector3){x, playfield->maxHeight - playfield->fieldSize.y * 0.5f, 0}, divWidth, 5, 5, rangeLimitColor);
    }
}

void PlayfieldFullDraw(Playfield *playfield, Camera3D camera)
{
    ClearBackground(COLOR_FROM_HEX(0x4488cc));

    BeginTextureMode(bubbleFieldTexture);
    rlSetClipPlanes(-128.0f, 128.0f);
    BeginMode3D(camera);

    ClearBackground(BLANK);
    PlayfieldDrawBubbles(playfield, camera);
    EndMode3D();
    EndTextureMode();

    float outlineSize = 1.0f;
    float outlineColor[4] = { 1.0f, 1.0f, 1.0f, 1.0f };
    float textureSize[2] = { (float)bubbleFieldTexture.texture.width, (float)bubbleFieldTexture.texture.height };

    SetShaderValue(bubbleOutlineShader, outlineSizeLoc, &outlineSize, SHADER_UNIFORM_FLOAT);
    SetShaderValue(bubbleOutlineShader, outlineColorLoc, outlineColor, SHADER_UNIFORM_VEC4);
    SetShaderValue(bubbleOutlineShader, textureSizeLoc, textureSize, SHADER_UNIFORM_VEC2);

    rlDisableDepthMask();
    BeginShaderMode(bubbleOutlineShader);
    DrawTexturePro(bubbleFieldTexture.texture, (Rectangle){0, 0, (float)bubbleFieldTexture.texture.width, 
        -(float)bubbleFieldTexture.texture.height}, (Rectangle){0, 0, (float)GetScreenWidth(), (float)GetScreenHeight()}, 
        (Vector2){0, 0}, 0.0f, WHITE);
    EndShaderMode();
    rlEnableDepthMask();

    BeginMode3D(camera);
    PlayfieldDrawRange(playfield, camera);
    EndMode3D();

    const char *difficultyText = "Tutorial";
    switch (gameDifficulty)
    {
    case GAME_DIFFICULTY_EASY: difficultyText = "Easy"; break;
    case GAME_DIFFICULTY_NORMAL: difficultyText = "Normal"; break;
    case GAME_DIFFICULTY_HARD: difficultyText = "Hard"; break;
    default:
        break;
    }
    const char *modeText = TextFormat("Mode: %s", difficultyText);
    int screenWidth = GetScreenWidth();
    int x = screenWidth - 220;
    DrawText2("Highscores", x, 25, WHITE);
    DrawText2(modeText, x, 55, WHITE);
    DifficultyScores table = storage.scores[gameDifficulty];
    for (int i = 0; i < 8; i++)
    {
        HighscoreEntry entry = table.highscores[i];
        if (entry.score == 0) break;
        char buffer[64];
        sprintf(buffer, "%d:", i + 1);
        int y = 110 + i * 30;
        DrawText2(buffer, x + 18 - MeasureText2(buffer), y, WHITE);
        sprintf(buffer, "%d", entry.score);
        DrawText2(buffer, x + 50 - MeasureText2(buffer) / 2, y, WHITE);
        sprintf(buffer, "%s", entry.date);
        DrawText2(buffer, screenWidth - 10 - MeasureText2(buffer), y, WHITE);
    }

    int sx = 0, sy = 0;
    if (playfield->comboPop > 2)
    {
        sx = 1;
    }
    if (playfield->strikes > 0)
    {
        sx = 0;
        sy = 1;
    }
    if (currentScene == GAME_SCENE_GAMEOVER)
    {
        sx = 1;
        sy = 1;
    }
    DrawTextureRec(texShrimpy, 
        (Rectangle){sx * 256, sy * 256, 256, 256},
        (Vector2){GetScreenWidth() - 200, GetScreenHeight() - 260}, WHITE);
}

void UpdateSceneGameOver()
{
    if (IsKeyPressed(KEY_ENTER))
    {
        nextScene = GAME_SCENE_MENU;
    }
    // Draw
    //----------------------------------------------------------------------------------

    ClearBackground(COLOR_FROM_HEX(0x4488cc));

    PlayfieldFixedUpdate(&playfield, 1);
    PlayfieldFullDraw(&playfield, camera);

    DrawText3("Game Over", 20, 20, WHITE);
    
    DrawText2(TextFormat("Final Score: %d", finalScores.score), 20, 90, WHITE);

    if (Button("< Menu", (Vector2){20, GetScreenHeight() - 110}, (Vector2){180, 90}))
    {
        nextScene = GAME_SCENE_MENU;
    }
}


void UpdateScenePlay()
{
    if (IsKeyPressed(KEY_ESCAPE))
    {
        nextScene = GAME_SCENE_MENU;
    }

    if (IsClickActioned() && playfield.strikes < MAX_STRIKES)
    {
        Vector2 pos = PlayfieldScreenToSpawnPosition(&playfield, camera, 
            GetMousePosition());
        if (pos.y >= 0.0f && pos.y <= playfield.fieldSize.y
            && pos.x >= 0.0f && pos.x <= playfield.fieldSize.x)
        {
            PlayfieldTryAddBubble(&playfield, pos);
        }
    }
    
    while (gameTime.fixedTime < gameTime.time)
    {
        gameTime.fixedTime += gameTime.fixedDeltaTime;
        PlayfieldFixedUpdate(&playfield, 0);
    }

    // Draw
    //----------------------------------------------------------------------------------

    PlayfieldFullDraw(&playfield, camera);

    PlayfieldScores scores = CalculatePlayfieldScores(&playfield);
    int y = 50;
    DrawText2("Shrimpy Bubble", 10, 20, WHITE);
    DrawText2("Merge", 150, 42, WHITE);
    
    DrawText2(TextFormat("Score: %d", scores.score), 10, y+=30, WHITE);
    DrawText2(TextFormat("Combos: %d", playfield.comboPop), 10, y+=30, WHITE);
    // DrawText2(TextFormat("Bubbles: %d", scores.bubbleCount), 10, y+=30, WHITE);
    // DrawText2(TextFormat("Spawned: %d", playfield.spawnedBubbleCount), 10, y+=30, WHITE);
    // DrawText2(TextFormat("Outside: %d", scores.outsideBubbleCount), 10, y+=30, WHITE);
    DrawText2(TextFormat("Strikes: %d", playfield.strikes), 10, y+=40, 
        playfield.strikes > 0 ? ColorLerp(WHITE, COLOR_FROM_HEX(0xff4433), sinf(gameTime.time * 10.0f)) : WHITE);
    if (playfield.strikes >= MAX_STRIKES)
    {
        DrawText2(TextFormat("Gameover in: %.1fs", GAMEOVER_COOLDOWN - playfield.gameoverCooldown), 10, y+=30, RED);
    }

    if (Button("< Menu", (Vector2){10, GetScreenHeight() - 100}, (Vector2){150, 80}))
    {
        nextScene = GAME_SCENE_MENU;
    }
}

void UpdateSceneSettings()
{
    ClearBackground(COLOR_FROM_HEX(0x4488cc));
    int hCenter = 200;

    DrawTextEx(fontMainLarge, "Shrimpy Bubble Merge Settings", 
        (Vector2) {50, 15}, fontMainLarge.baseSize, 1, WHITE);
    
    DrawTextureRec(texShrimpy, 
        (Rectangle){0, 256, 256, 256},
        (Vector2){GetScreenWidth() - 400, GetScreenHeight() - 260}, WHITE);

    int buttonH = 60;
    int y = 50;
    
    if (Button(playSounds ? "Sound: On" : "Sound: Off", 
        (Vector2){hCenter - 100, y += 60}, (Vector2){200, buttonH}))
    {
        playSounds = !playSounds;
        SaveStorage();
    }

    if (Button("Clear data", (Vector2){hCenter - 100, y += 60}, (Vector2){200, buttonH}))
    {
        nextScene = GAME_SCENE_MENU;
        memset(&storage, 0, sizeof(storage));
        SaveStorage();
    }

    if (Button("< Menu", (Vector2){hCenter - 100, y += 60}, (Vector2){200, buttonH}))
    {
        nextScene = GAME_SCENE_MENU;
    }
}

void UpdateSceneMenu()
{
    ClearBackground(COLOR_FROM_HEX(0x4488cc));

    int hCenter = 200;

    DrawTextEx(fontMainLarge, "Shrimpy Bubble Merge", 
        (Vector2) {50, 15}, fontMainLarge.baseSize, 1, WHITE);
    
    DrawTextureRec(texShrimpy, 
        (Rectangle){0, 0, 256, 256},
        (Vector2){GetScreenWidth() - 400, GetScreenHeight() - 260}, WHITE);

    int buttonH = 60;
    int y = 100;
    if (Button("Tutorial", (Vector2){hCenter - 100, y}, (Vector2){200, buttonH}))
    {
        gameDifficulty = GAME_DIFFICULTY_TUTORIAL;
        nextScene = GAME_SCENE_PLAY;
    }

    if (Button("Play easy", (Vector2){hCenter - 100, y += 60}, (Vector2){200, buttonH}))
    {
        gameDifficulty = GAME_DIFFICULTY_EASY;
        nextScene = GAME_SCENE_PLAY;
    }

    if (Button("Play normal", (Vector2){hCenter - 100, y+=60}, (Vector2){200, buttonH}))
    {
        gameDifficulty = GAME_DIFFICULTY_NORMAL;
        nextScene = GAME_SCENE_PLAY;
    }

    if (Button("Play hard", (Vector2){hCenter - 100, y+=60}, (Vector2){200, buttonH}))
    {
        gameDifficulty = GAME_DIFFICULTY_HARD;
        nextScene = GAME_SCENE_PLAY;
    }

    if (Button("Settings", (Vector2){hCenter - 100, y+=60}, (Vector2){200, buttonH}))
    {
        nextScene = GAME_SCENE_SETTINGS;
    }
}
#if defined(PLATFORM_WEB)
#include <emscripten.h>

// Function to store data in Local Storage
void StoreData(const char *key, const char *value) {
    EM_ASM_({
        localStorage.setItem(UTF8ToString($0), UTF8ToString($1));
    }, key, value);
}

// Function to retrieve data from Local Storage
const char* RetrieveData(const char *key) {
    return (const char*)EM_ASM_INT({
        var value = localStorage.getItem(UTF8ToString($0));
        if (value === null) {
            return 0;
        }
        var lengthBytes = lengthBytesUTF8(value) + 1;
        var stringOnWasmHeap = _malloc(lengthBytes);
        stringToUTF8(value, stringOnWasmHeap, lengthBytes);
        return stringOnWasmHeap;
    }, key);
}
#else
void StoreData(const char *key, const char *value) {}
const char* RetrieveData(const char *key) { return 0; }
#endif

uint32_t RetrieveUInt32(const char *key)
{
    const char *value = RetrieveData(key);
    if (value)
    {
        uint32_t result = atoi(value);
        free((void*)value);
        return result;
    }
    return 0;
}

void StoreUInt32(const char *key, uint32_t value)
{
    char buffer[16];
    sprintf(buffer, "%d", value);
    StoreData(key, buffer);
}

void RetrieveFixedChar(const char *key, char *buffer, int size)
{
    const char *value = RetrieveData(key);
    if (value)
    {
        strncpy(buffer, value, size);
        free((void*)value);
    }
    else
    {
        buffer[0] = '\0';
    }
}

void StoreFixedChar(const char *key, const char *value)
{
    StoreData(key, value);
}


void LoadStorage()
{
    playSounds = RetrieveUInt32("storage.playSounds");
    // ignore version as this is first version. Upgrades need to be backwards compatible
    for (int i = 0; i < 4; i++)
    {
        for (int j = 0; j < 8; j++)
        {
            char key[64];
            sprintf(key, "storage.highscore_%d_%d", i, j);
            storage.scores[i].highscores[j].score = RetrieveUInt32(key);
            sprintf(key, "storage.name_%d_%d", i, j);
            RetrieveFixedChar(key, storage.scores[i].highscores[j].name, 16);
            sprintf(key, "storage.date_%d_%d", i, j);
            RetrieveFixedChar(key, storage.scores[i].highscores[j].date, 16);
        }
    }

    storage.tutorialStep = RetrieveUInt32("storage.tutorialStep");
    storage.tutorialStepCount = RetrieveUInt32("storage.tutorialStepCount");
    storage.tutorialCompleted = RetrieveUInt32("storage.tutorialCompleted");
    storage.startups = RetrieveUInt32("storage.startups");
    StoreUInt32("storage.startups", storage.startups + 1);
}

void SaveStorage()
{
    StoreUInt32("storage.playSounds", playSounds);
    StoreUInt32("storage.version", 1);
    for (int i = 0; i < 4; i++)
    {
        for (int j = 0; j < 8; j++)
        {
            char key[64];
            sprintf(key, "storage.highscore_%d_%d", i, j);
            StoreUInt32(key, storage.scores[i].highscores[j].score);
            sprintf(key, "storage.name_%d_%d", i, j);
            StoreFixedChar(key, storage.scores[i].highscores[j].name);
            sprintf(key, "storage.date_%d_%d", i, j);
            StoreFixedChar(key, storage.scores[i].highscores[j].date);
        }
    }
    StoreUInt32("storage.tutorialStep", storage.tutorialStep);
    StoreUInt32("storage.tutorialStepCount", storage.tutorialStepCount);
    StoreUInt32("storage.tutorialCompleted", storage.tutorialCompleted);
}

void LogStorage()
{
    TraceLog(LOG_INFO, "[storage] Storage log");
    for (int i = 0; i < 4; i++)
    {
        for (int j = 0; j < 8; j++)
        {
            TraceLog(LOG_INFO, "[storage] Highscore %d %d: %d %s %s", i, j, storage.scores[i].highscores[j].score, 
                storage.scores[i].highscores[j].name, storage.scores[i].highscores[j].date);
        }
    }
    TraceLog(LOG_INFO, "[storage] Tutorial: %d %d %d", storage.tutorialStep, storage.tutorialStepCount, storage.tutorialCompleted);
    TraceLog(LOG_INFO, "[storage] Startups: %d", storage.startups);
}

#include <time.h>
void StoreHighScore(PlayfieldScores scores)
{
    DifficultyScores *difficultyScores = &storage.scores[gameDifficulty];

    for (int i = 0; i < 8; i++)
    {
        if (scores.score > difficultyScores->highscores[i].score)
        {
            for (int j = 7; j > i; j--)
            {
                difficultyScores->highscores[j] = difficultyScores->highscores[j - 1];
            }
            difficultyScores->highscores[i].score = scores.score;
            time_t now = time(0);
            struct tm *tm = localtime(&now);
            strftime(difficultyScores->highscores[i].date, 16, "%Y-%m-%d", tm);
            difficultyScores->highscores[i].name[0] = '\0';
            SaveStorage();
            break;
        }
    }
}

void TutorialBubble(int step, const char *text, int x, int y, int width, int height)
{
    if (playfield.tutorialStep != step) return;
    Rectangle rect = {x, y, width, height};
    DrawTextureNPatch(texUiAtlas, nPatchButton, rect, Vector2Zero(), 0.0f, COLOR_FROM_HEX(0xcceeff));
    
    Font font = fontMainMedium;
    int lineHeight = font.baseSize - 4;
    int lineCount = 1;
    for (int i = 0; i < strlen(text); i++)
    {
        if (text[i] == '\n')
        {
            lineCount++;
        }
    }
    int yTop = y;
    float textHeight = lineCount * lineHeight;
    const char *line = text;
    y = y + height * 0.5f - textHeight * 0.5f;
    while (*line) {
        char lineBuffer[256];
        const char *end = line;
        while (*end && *end != '\n') {
            lineBuffer[end - line] = *end;
            end++;
        }
        lineBuffer[end - line] = '\0';
        
        Vector2 tsize = MeasureTextEx(font, lineBuffer, font.baseSize, 1);
        DrawTextEx(font, lineBuffer, (Vector2){ x + width * 0.5f - tsize.x * 0.5f, y }, font.baseSize, 1, BLACK);

        line = end;
        while (*line == '\n') line++;
        y += lineHeight;
    }
    
    DrawTextureRec(texUiAtlas, rectUiAtlasCircle, (Vector2){x + width - 30, yTop}, COLOR_FROM_HEX(0xccddff));
    DrawTextureRec(texUiAtlas, rectUiAtlasShrimpy, (Vector2){x + width, yTop + 10}, WHITE);
}

void TutorialSetClicksBlocked(int step, int blocked)
{
    if (playfield.tutorialStep != step) return;
    isClickActionBlocked = blocked;
}

void TutorialProceedOnClick(int step)
{
    if (playfield.tutorialStep != step) return;
    if (IsMouseButtonReleased(MOUSE_LEFT_BUTTON)) {
        playfield.nextTutorialStep = step + 1;
    }
}

void TutorialHighlightCircle(int step, int x, int y, int radius)
{
    if (playfield.tutorialStep != step) return;
    float pulse = sinf(gameTime.time * 10.0f) * 0.25f + 1.0f;
    DrawCircleLines(x, y, radius * pulse, COLOR_FROM_HEX(0xff0000));
    DrawCircleLines(x, y, radius * pulse + 0.5f, COLOR_FROM_HEX(0xff0000));
    DrawCircleLines(x, y, radius * pulse + 1.0f, COLOR_FROM_HEX(0xff0000));
}

void TutorialSpawnBubble(int step, int x, int y, int level, int type, int nextLevel, int nextType)
{
    if (playfield.tutorialStep != step || !playfield.tutorialStepStart) return;
    Vector2 pos = PlayfieldScreenToSpawnPosition(&playfield, camera, 
            (Vector2){x, y});
    playfield.nextBubbleLevel = level;
    playfield.nextBubbleType = type;
    PlayfieldTryAddBubble(&playfield, pos);
    playfield.nextBubbleLevel = nextLevel;
    playfield.nextBubbleType = nextType;
}

void UpdateTutorialSystem_PlayTutorial()
{
    int step = 0;
    int boxHeight = 105;   
    TutorialSetClicksBlocked(step, 1);
    TutorialBubble(step, "Welcome to my fishtank...", 10, 10, 600, boxHeight);
    TutorialProceedOnClick(step);
    step++;

    TutorialSetClicksBlocked(step, 1);
    TutorialBubble(step, "It is very boring here.\nSo I invented this little game:", 10, 10, 600, boxHeight);
    TutorialProceedOnClick(step);
    step++;

    TutorialSetClicksBlocked(step, 1);
    TutorialBubble(step, "Please tap on the marked area ...", 10, 10, 600, boxHeight);
    TutorialProceedOnClick(step);
    TutorialHighlightCircle(step, GetScreenWidth() / 2, GetScreenHeight() - 50, 20);
    step++;

    TutorialSpawnBubble(step, GetScreenWidth() / 2, GetScreenHeight() - 50, 1, 0, 1, 0);
    TutorialBubble(step, "Very good! Now do it again!", 10, 10, 600, boxHeight);
    TutorialProceedOnClick(step);
    TutorialHighlightCircle(step, GetScreenWidth() / 2, GetScreenHeight() - 50, 20);
    TutorialSetClicksBlocked(step, 1);
    step++;
    
    TutorialSpawnBubble(step, GetScreenWidth() / 2, GetScreenHeight() - 50, 1, 0, 2, 1);
    TutorialBubble(step, 
        "These two bubbles have the same size and\n"
        "color. When they touch, they will merge.", 10, 10, 600, boxHeight);
    TutorialProceedOnClick(step);
    TutorialSetClicksBlocked(step, 1);
    step++;
    
    TutorialBubble(step, "But currently, they don't :(\nLet me help you a little.", 10, 10, 600, boxHeight);
    TutorialProceedOnClick(step);
    TutorialSetClicksBlocked(step, 1);
    step++;
    TutorialBubble(step, "...", 10, 10, 600, boxHeight);
    if (step == playfield.tutorialStep) {
        playfield.enableTutorialMerge = 1;
        if (playfield.bubbles[0].bubbleLevel == 2)
        {
            playfield.nextTutorialStep = step + 1;
        }
    }
    step++;
    TutorialBubble(step, "See? The two bubbles merged and now\nit is bigger - look and the two dots it got.", 10, 10, 600, boxHeight);
    TutorialProceedOnClick(step);
    TutorialSetClicksBlocked(step, 1);
    step++;
    TutorialBubble(step, "Oh, the next bubble has a different color.\nTap again!", 10, 10, 600, boxHeight);
    TutorialProceedOnClick(step);
    TutorialHighlightCircle(step, GetScreenWidth() / 2, GetScreenHeight() - 50, 20);
    TutorialSetClicksBlocked(step, 1);
    
    step++;
    TutorialSpawnBubble(step, GetScreenWidth() / 2, GetScreenHeight() - 50, 2, 1, 2, 1);
    TutorialBubble(step, "It won't merge with the bright bubble.\nI don't know why.", 10, 10, 600, boxHeight);
    TutorialProceedOnClick(step);
    
    step++;
    TutorialBubble(step, "But look! We have another bubble of \nthe same color! Release it!", 10, 10, 600, boxHeight);
    TutorialProceedOnClick(step);
    TutorialHighlightCircle(step, GetScreenWidth() / 2, GetScreenHeight() - 50, 20);
    step++;
    TutorialSpawnBubble(step, GetScreenWidth() / 2, GetScreenHeight() - 50, 2, 1, 2, 1);
    TutorialBubble(step, "Wait for it...", 10, 10, 600, boxHeight);
    if (step == playfield.tutorialStep) {
        playfield.enableTutorialMerge = 1;
        if (playfield.bubbles[1].bubbleLevel == 3)
        {
            playfield.nextTutorialStep = step + 1;
        }
    }

    step++;
    TutorialBubble(step, "Just watch out: When the bubbles go too high\nor low, they all pop :(", 10, 10, 600, boxHeight);
    TutorialProceedOnClick(step);
    
    step++;
    TutorialBubble(step, "I love these bubbles!\nGo and try to see how many you can get!", 10, 10, 600, boxHeight);
    TutorialProceedOnClick(step);
    if (step == playfield.tutorialStep) {
        playfield.enableTutorialMerge = 0;
    }
    
    step++;
    TutorialSetClicksBlocked(step, 0);

    if (playfield.strikes > 0 && playfield.tutorialStep == step)
    {
        playfield.nextTutorialStep = step + 1;
    }
    step++;

    TutorialBubble(step, "Oh no! The bubbles are about to pop!\nCareful, you have 2 more tries!", 10, 10, 600, boxHeight);
    TutorialProceedOnClick(step);
    TutorialSetClicksBlocked(step, 1);
    step++;
    TutorialSetClicksBlocked(step, 0);
}

void UpdateTutorialSystem()
{
    if (currentScene == GAME_SCENE_PLAY && gameDifficulty == GAME_DIFFICULTY_TUTORIAL)
    {
        UpdateTutorialSystem_PlayTutorial();
    }

    if (playfield.nextTutorialStep != playfield.tutorialStep)
    {
        playfield.tutorialStepStart = 1;
    }
    else
    {
        playfield.tutorialStepStart = 0;
    }
    playfield.tutorialStep = playfield.nextTutorialStep;
}

// out of time
// void DrawWaterPlant(float x, float y, float height, Color color)
// {
//     Vector2 lineStrip[90];
//     float yStep = height / 30;
//     for (int i = 0; i < 45; i++)
//     {
//         float t = i / 45.0f;
//         float w = powf(t, 0.6f) * height / 90.0f;
//         float offset = sinf(gameTime.time * 2.0f + i * 0.1f) * 2.0f;
//         lineStrip[i * 2] = (Vector2){x + w + offset, y - i * yStep};
//         lineStrip[i * 2 + 1] = (Vector2){x - w + offset, y - i * yStep};
//     }
//     DrawTriangleStrip(lineStrip, 90, color);
// }

int main(void)
{
    // Initialization
    //--------------------------------------------------------------------------------------
    const int screenWidth = 800;
    const int screenHeight = 450;

    InitWindow(screenWidth, screenHeight, "Shrimpy Bubble Merge");
    InitAudioDevice();

    SetTargetFPS(60);               // Set our game to run at 60 frames-per-second
    //--------------------------------------------------------------------------------------

    TraceLog(LOG_INFO, "loading shaders");
    bubbleOutlineShader = LoadShader(0, "data/bubble_outline.fs");
    // Get shader locations
    outlineSizeLoc = GetShaderLocation(bubbleOutlineShader, "outlineSize");
    outlineColorLoc = GetShaderLocation(bubbleOutlineShader, "outlineColor");
    textureSizeLoc = GetShaderLocation(bubbleOutlineShader, "textureSize");

    LoadStorage();
    LogStorage();

    texShrimpy = LoadTexture("data/shrimpy.png");
    texLeftCorner = LoadTexture("data/left-corner.png");
    texUiAtlas = LoadTexture("data/ui-atlas.png");
    texGround = LoadTexture("data/ground.png");
    fontMainMedium = LoadFontEx("data/IndieFlower-Regular.ttf", 40, 0, 0);
    fontMainLarge = LoadFontEx("data/IndieFlower-Regular.ttf", 70, 0, 0);
    sfxPop = LoadSound("data/pop1.wav");
    bubbleModel = LoadModelFromMesh(GenMeshSphere(1.0f, 4, 24));

    Vector2 floatingBubbles[32];
    Vector2 floatingBubblesVelocities[32];
    for (int i = 0; i < 32; i++)
    {
        floatingBubbles[i] = (Vector2){GetRandomValue(0, screenWidth), GetRandomValue(0, screenHeight)};
        floatingBubblesVelocities[i] = (Vector2){GetRandomValue(-100, 100) * 0.1f, GetRandomValue(-100, 100) * 0.1f};
    }
    // Main game loop
    while (!WindowShouldClose())    // Detect window close button or ESC key
    {
        if (bubbleFieldTexture.texture.width != GetScreenWidth() || bubbleFieldTexture.texture.height != GetScreenHeight())
        {
            UnloadRenderTexture(bubbleFieldTexture);
            bubbleFieldTexture = LoadRenderTexture(GetScreenWidth(), GetScreenHeight());
        }

        float dt = GetFrameTime();
        // clamp dt to prevent large time steps, e.g. when browser tab is inactive
        if (dt > 0.2f) dt = 0.2f;
        gameTime.time += dt;
        gameTime.deltaTime = dt;

        BeginDrawing();

        DrawTexture(texLeftCorner, 0, GetScreenHeight() - texLeftCorner.height - 16, WHITE);

        float decay = expf(-dt * 2.0f);
        for (int i=0; i < 32; i++)
        {
            DrawCircle(floatingBubbles[i].x, floatingBubbles[i].y, i %4 + 2, ColorLerp(WHITE, COLOR_FROM_HEX(0x4488cc), sinf(gameTime.time + i * 0.1f) * 0.5f + 0.5f));
            floatingBubbles[i] = Vector2Add(floatingBubbles[i], Vector2Scale(floatingBubblesVelocities[i], dt));
            floatingBubblesVelocities[i] = Vector2Add(floatingBubblesVelocities[i], (Vector2){GetRandomValue(-100,100) * 0.05f, -10.0f * dt});
            floatingBubblesVelocities[i] = Vector2Scale(floatingBubblesVelocities[i], decay);
            if (floatingBubbles[i].x < 0 || floatingBubbles[i].x > GetScreenWidth()
                || floatingBubbles[i].y < 0 || floatingBubbles[i].y > GetScreenHeight())
            {
                floatingBubbles[i].x = GetRandomValue(0, GetScreenWidth());
                floatingBubbles[i].y = GetRandomValue(0, GetScreenHeight());
                floatingBubblesVelocities[i].x = GetRandomValue(-100, 100) * 0.1f;
                floatingBubblesVelocities[i].y = GetRandomValue(-100, 100) * 0.1f;
            }
        }

        int y = GetScreenHeight() - texGround.height;
        for (int x=0; x<screenWidth; x+=texGround.width)
        {
            DrawTexture(texGround, x, y, WHITE);
        }

        // draw wavy water line at top
        Vector2 waterLine[80];
        float waterXStep = GetScreenWidth() / 38;
        for (int i = 0; i < 40; i++)
        {
            waterLine[i * 2] = (Vector2){i * waterXStep, 0};
            waterLine[i * 2 + 1] = (Vector2){i * waterXStep, 15
                + sinf(gameTime.time * 2.9f + waterLine[i * 2 + 1].x * 0.008f) * 2.0f
                + sinf(gameTime.time * 4.3f + waterLine[i * 2 + 1].x * 0.015f) * 1.0f
                + sinf(gameTime.time * 6.9f + waterLine[i * 2 + 1].x * 0.032f) * 0.5f
            };
        }
        DrawTriangleStrip(waterLine, 80, COLOR_FROM_HEX(0x88ccff));
        
        for (int i = 0; i < 40; i++)
        {
            float x = i * waterXStep;
            waterLine[i * 2 + 1] = (Vector2){x, 16
                + sinf(gameTime.time * 2.9f + x * 0.008f) * 2.0f
                + sinf(gameTime.time * 4.3f + x * 0.015f) * 1.0f
                + sinf(gameTime.time * 6.9f + x * 0.032f) * 0.5f
            };
            waterLine[i * 2] = (Vector2){x, waterLine[i * 2 + 1].y - 2.0f};
        }
        DrawTriangleStrip(waterLine, 80, COLOR_FROM_HEX(0xeef8ff));

        // DrawWaterPlant(50, GetScreenHeight() - 20, 130, COLOR_FROM_HEX(0x449955));
        
        

        switch (currentScene)
        {
            default: UpdateSceneMenu(); break;
            case GAME_SCENE_SETTINGS: UpdateSceneSettings(); break;
            case GAME_SCENE_PLAY: UpdateScenePlay(); break;
            case GAME_SCENE_GAMEOVER: UpdateSceneGameOver(); break;
        }

        UpdateTutorialSystem();

        switch (nextScene)
        {
            case GAME_SCENE_NONE: break;
            default: currentScene = nextScene; break;
            case GAME_SCENE_GAMEOVER:
                StoreHighScore(CalculatePlayfieldScores(&playfield));
                currentScene = GAME_SCENE_GAMEOVER;
                break;
            case GAME_SCENE_PLAY: 
                playfield = (Playfield){
                    .fieldSize = {230, 300},
                    .waterLevel = 200.0f,
                    .minHeight = 90.0f,
                    .maxHeight = 280.0f,
                };
                if (gameDifficulty == GAME_DIFFICULTY_TUTORIAL)
                {
                    playfield.waterLevel = 180.0f;
                }
                currentScene = GAME_SCENE_PLAY; break;
        }

        if (nextScene != GAME_SCENE_NONE)
        {
            gameTime.fixedTime = gameTime.time;
            isClickActionBlocked = 0;
            playfield.tutorialStep = 0;
            playfield.nextTutorialStep = 0;
        }

        nextScene = GAME_SCENE_NONE;
        EndDrawing();
    }

    // De-Initialization
    //--------------------------------------------------------------------------------------
    CloseWindow();        // Close window and OpenGL context
    //--------------------------------------------------------------------------------------

    return 0;
}